from automate_office.automate_word.MyWordCreator import WordCreator
from docx.oxml.ns import qn
from docx.enum.section import WD_SECTION_START, WD_ORIENT

from automate_office.automate_word.helper_func import display_paragraphs, add_font_style, add_paragraph_style
from automate_office.automate_word.settings import DOC_FONT_SETTINGS

if __name__ == "__main__":
    my_word = WordCreator()

    p_s_1 = add_paragraph_style(my_word.doc, style_name="1", font_info=("微软雅黑", 12, False, False, False, (0, 0, 0)), first_line_indent=-1, line_spacing=0.5)
    p_s_2 = add_paragraph_style(my_word.doc, style_name="2", font_info=("微软雅黑", 16, False, False, False, (0, 0, 0)), first_line_indent=-4, line_spacing=4)


    my_word.add_paragraph(
        paragraph_info=[
            ["a1", None],
            ["a2", ("微软雅黑", 12, False, True, False, (255, 0, 0))],
            ["a3", None],
            ["a4", ("微软雅黑", 12, True, True, False, (0, 0, 255))],
            ["a5", None],
            ["a6" * 100, None],
        ],
        paragraph_style=p_s_1
    )

    my_word.add_paragraph(
        paragraph_info=[
            ["a1", ("微软雅黑", 12, True, False, False, (0, 0, 0))],
            ["a2", None],
            ["a3", ("微软雅黑", 12, False, False, True, (0, 255, 0))],
            ["a4", None],
            ["a5", ("微软雅黑", 12, False, True, True, (255, 255, 0))],
            ["a6" * 100, None],
        ],
        paragraph_style=p_s_2
    )
    my_word.save_doc()






from docx.shared  import Inches


DOC_FONT_SETTINGS = {
    "BOOL_SETTINGS": [
        # 加粗
        "bold", 
        # 斜体
        "italic", 
        # 下划线
        "underline", 
        # 以下的没看出效果，先列出来
        "double_strike", 
        "emboss", 
        "hidden", 
        "imprint", 
        "outline", 
        "rtl", 
        "shadow", 
        "snap_to_grid", 
        "spec_vanish", 
        "strike", 
        "subscript", 
        "superscript", 
        "web_hidden"
    ]

}


import re
import pathlib

from openpyxl import Workbook, worksheet, load_workbook
from openpyxl.styles import PatternFill, Border, Side, Alignment, Protection, Font, colors
from openpyxl.styles.alignment import horizontal_alignments, vertical_aligments


def create_new_excel():
    # 实例化
    wb = Workbook()
    return wb


def open_excel(filename):
    return load_workbook(filename=filename)


def save_excel(wb: Workbook, save_path=None):
    if save_path is None:
        save_path = pathlib.Path.home().joinpath("Downloads/output.xlsx")
    wb.save(save_path)


def create_new_sheet(wb: Workbook, sheet_name, index=None):
    return wb.create_sheet(
        title=sheet_name,
        # None代表在末尾加，否则就是index位置添加
        index=index
    )


def remove_sheet(wb: Workbook, sheet_name=None, sheet_index=None):
    if sheet_name is not None:
        wb.remove_sheet(worksheet=wb.get_sheet_by_name(sheet_name))
    elif sheet_index is not None:
        wb.remove_sheet(wb.worksheets[sheet_index])
    else:
        raise ValueError("sheet_name和sheet_index至少要有一个不为None")


def rename_sheet(ws: worksheet.worksheet.Worksheet, sheet_name):
    ws.title = sheet_name


def set_sheet_color(ws: worksheet.worksheet.Worksheet, aRGB ="00000000"):
    ws.sheet_properties.tabColor = colors.RgbColor(aRGB)


def write_data(ws: worksheet.worksheet.Worksheet, data):
    for row in data:
        ws.append(row)


def create_font_style(font_name="微软雅黑", font_size=12, font_color="00000000", bold=False, italic=False, vert_align=None, underline="none", strike=False):
    """ cell.font = return
        vertAlign: 垂直对齐方式, 'superscript', 'subscript', 'baseline', 前两个是上下标, 最后一个看不出来, 和正常的不一样
        underline: 下划线, 'single', 'double', 'singleAccounting', 'doubleAccounting', single是单下划线, double是双的, 带Accounting的比不带的要高一些, 估计是会计用的吧
        strike: 删除线
        font_color: aRGB, 十六进制
    """
    return Font(
        name=font_name,
        size=font_size,
        bold=bold,
        italic=italic,
        vertAlign=vert_align,
        underline=underline,
        strike=strike,
        color=font_color
    )


def create_fill_style(fill_type="none", start_color="FFFFFFFF", end_color="FF000000"):
    """ 单元格填充颜色, cell.fill = return
    Args:
        fill_type (_type_, optional): 各种花纹, 感觉solid就够用: ["none", "solid", "darkDown", "darkGray", "darkGrid", "darkHorizontal", "darkTrellis", "darkUp", "darkVertical", "gray0625", "gray125", "lightDown", "lightGray", "lightGrid", "lightHorizontal", "lightTrellis", "lightUp", "lightVertical", "mediumGray"]
        start_color (str, optional): 默认是start_color, 但是有花纹的话end_color就用上了
        end_color (str, optional): 
    """
    __fill_types__ = ["none", "solid", "darkDown", "darkGray", "darkGrid", "darkHorizontal", "darkTrellis", "darkUp", "darkVertical", "gray0625", "gray125", "lightDown", "lightGray", "lightGrid", "lightHorizontal", "lightTrellis", "lightUp", "lightVertical", "mediumGray"]
    if fill_type is not None:
        assert fill_type in __fill_types__, f"style应该是{', '.join(__fill_types__)}中的一种"

    return PatternFill(
        fill_type=fill_type,
        start_color=start_color,
        end_color=end_color
    )


def create_border_style(
        left_style=None, right_style=None, top_style=None, bottom_style=None, color="00000000"
        # 这些参数没看到效果先忽略
        # diagonal_style, diagonal_direction,
        # outline_style, vertical_style, horizontal_style
    ):
    """ cell.border = return
    style:  各种框框类型, 常用的是dashed/thin/thick, 'dashDot','dashDotDot', 'dashed','dotted', 'double','hair', 'medium', 'mediumDashDot', 'mediumDashDotDot', 'mediumDashed', 'slantDashDot', 'thick', 'thin'
    Args:
        left_style (_type_): _description_
        right_style (_type_): _description_
        top_style (_type_): _description_
        bottom_style (_type_): _description_
        diagonal_style (_type_): _description_
        diagonal_direction (_type_): _description_
        outline_style (_type_): _description_
        vertical_style (_type_): _description_
        horizontal_style (_type_): _description_

    Returns:
        _type_: _description_
    """

    __border_types__ = ('dashDot','dashDotDot', 'dashed','dotted', 'double','hair', 'medium', 'mediumDashDot', 'mediumDashDotDot', 'mediumDashed', 'slantDashDot', 'thick', 'thin')
    for cur_style in [left_style, right_style, top_style, bottom_style]:
        if cur_style is not None:
            assert cur_style in __border_types__, f"style应该是{', '.join(__border_types__)}中的一种"
    
    return Border(
        left=Side(border_style=left_style, color=color),
        right=Side(border_style=right_style, color=color),
        top=Side(border_style=top_style, color=color),
        bottom=Side(border_style=bottom_style, color=color)
        # diagonal=Side(border_style=None, color='FF000000'),
        # diagonal_direction=0, 
        # outline=Side(border_style=None, color='FF000000'),
        # vertical=Side(border_style=None, color='FF000000'),
        # horizontal=Side(border_style=None, color='FF000000')
    )


def creaete_alignment_style(horizontal="general", vertical="center", text_rotation=0, wrap_text=True, shrink_to_fit=False, ident=0):
    """ cell.alignment = retrn
    
        horizontal_alignments = ("general", "left", "center", "right", "fill", "justify", "centerContinuous", "distributed", )
        vertical_aligments = ("top", "center", "bottom", "justify", "distributed")
    Args:
        horizontal (str, optional): 水平
        vertical (str, optional): 垂直
        text_rotation (int, optional): 旋转角度
        wrap_text (bool, optional): _description_. 换行
        shrink_to_fit (bool, optional): 似乎没用，是否缩小适应单元格
        ident (int, optional): 似乎没用

    """
    if horizontal is not None:
        assert horizontal in horizontal_alignments, f"horizontal应该是{', '.join(horizontal_alignments)}中的一种"
    if vertical is not None:
        assert vertical in vertical_aligments, f"vertical应该是{', '.join(vertical_aligments)}中的一种"
    
    return Alignment(
        horizontal=horizontal,
        vertical=vertical,
        text_rotation=text_rotation,
        wrap_text=wrap_text,
        shrink_to_fit=shrink_to_fit,
        indent=ident
    )


def create_protection_style(locked=False, hidden=False):
    """
        锁定: 是不让改
        隐藏: 格式设置成";;;"之后，隐藏状态下公式编辑栏也看不到这个

        但是感觉隐藏没有卵用, 写个公式就能找出来了
    """
    return Protection(
        locked=locked,
        hidden=hidden
    )


def create_number_format():
    """
        Excel的数据格式, 比如百分比就是0.00%, 两位小数就是0.00等等
    """
    return "General"


def get_autojust_col_width(ws, n_cols):
    for col_idx in range(n_cols):
        col_dimension = ws.column_dimensions[chr(65 + col_idx)]
        col_dimension.width += 2


def get_autojust_col_width_bak(ws, start_row_idx, end_row_idx, col_idx=1):
    """本来用这个计算列的最大长度来控制列宽，但后来发现有自适应的方法

        Args:
            cur_str (_type_): _description_

        Returns:
            _type_: _description_
    """
    def get_length(cur_str):

        cur_str = str(cur_str)

        # 提取双字节字符的正则模板
        p = re.compile('[^\x00-\xff]+')  
        # 双字节字符数
        dualByteNum = len(''.join(p.findall(cur_str)))
        # 单字节字符数
        singleByteNum = len(cur_str) - dualByteNum
        
        return dualByteNum * 2 + singleByteNum

    col_max_length = 0

    for cur_row_idx in range(start_row_idx, end_row_idx + 1, 1):
        cur_cell = ws.cell(cur_row_idx, col_idx)
        cur_length = get_length(cur_cell.value)
        col_max_length = max(col_max_length, cur_length)


    return col_max_length + 2


    



from docx.shared import Pt, Inches, RGBColor
from docx.enum.style import WD_STYLE_TYPE

from automate_office.automate_word.settings import DOC_FONT_SETTINGS


def set_font(cur_font, cur_font_info):
    """设置字体
    Args:
        cur_font (_type_): pptx.text.text.font类型
        cur_font_info (_type_): tuple, 字体、字号、加粗、斜体、颜色
    """
    font_name, font_size, bold, italic, underline, rgb = cur_font_info

    cur_font.name = font_name
    cur_font.size = Pt(font_size)
    cur_font.bold = bold
    cur_font.italic = italic
    cur_font.underline = underline
    cur_font.color.rgb = RGBColor(*rgb)


def add_font_style(document, style_name, font_name, font_size=12, font_color=(0, 0, 0), bold=False, italic=False, underline=False):
    # CHARACTER STYLE似乎现在不能生效
    styles = document.styles
    # 新建一个叫Citation的style
    style = styles.add_style(style_name, WD_STYLE_TYPE.CHARACTER)
    set_font(font=style.font, font_name=font_name, font_size=font_size, font_color=font_color, bold=bold, italic=italic, underline=underline)


def add_paragraph_style(document, style_name, font_info, first_line_indent, line_spacing):
    styles = document.styles
    style = styles.add_style(style_name, WD_STYLE_TYPE.PARAGRAPH)
    # 设置字体
    set_font(style.font, font_info)
    # 设置段落
    # paragraph_format.space_before = Pt(18)
    # paragraph_format.space_after = Pt(12)
    paragraph_format = style.paragraph_format
    paragraph_format.first_line_indent = Inches(first_line_indent)
    paragraph_format.line_spacing = Inches(line_spacing)

def add_paragraph(cur_paragraph, paragraph_info):

    for text, cur_font_info in paragraph_info:
        run = cur_paragraph.add_run(text)
        if cur_font_info is not None:
            set_font(run.font, cur_font_info)


def display_paragraphs(cur_doc):

    for i, p in enumerate(cur_doc.paragraphs):
        print(i, p.text)



"""
    PPT版式(宽:高):
        - "4:3" : "25.4cm:19.05cm, 10英寸:7.5英寸",
        - "16:9": "33.87cm:19.05cm, 10英寸:7.5英寸",
        - "A4"  : "29.7cm:21cm, 11.693英寸:8.268英寸",
"""
import pathlib
from itertools import product

import pandas as pd
from pptx import Presentation, slide
from pptx.util import Inches, Cm, Pt
from pptx.enum.text import PP_ALIGN, MSO_ANCHOR
from pptx.dml.color import RGBColor
from pptx.enum.shapes import MSO_SHAPE
from pptx.chart.data import CategoryChartData
from pptx.enum.chart import XL_LEGEND_POSITION, XL_MARKER_STYLE, XL_LABEL_POSITION

from automate_office.automate_ppt.ppt_enums import PARAGRAPH_ALIGN, DEFAULT_PARAGRAPH_SPACING
from automate_office.automate_ppt.Charts import ChartCreator, set_color, set_font


class PPTCreator(object):

    def __init__(self, size_params=None):
        self._create_pptx(size_params)

    def _get_size(self, size_params):
        """
            size_params: 字典, {"unit_type": "inches", "width": 10, "height": 5.625}
        """
        if size_params is None:
            size_params = {
                "unit_type": "inches", 
                "width": 40/3, 
                "height": 7.5
            }
        if isinstance(size_params, dict):
            unit_type = {
                "inches": Inches,
                "cm": Cm
            }.get(
                size_params.get("unit_type", "inches"), Inches
            )

            self.height = unit_type(size_params.get("height", 7.5))
            self.width = unit_type(size_params.get("width", 40/3))

        else:
            raise ValueError("size_params要求是字典格式")

    @staticmethod
    def get_absolute_cordinate(length, rel_ratio):
        return Inches(length.inches * rel_ratio)

    def _create_pptx(self, size_params):
        self._get_size(size_params=size_params)
        self.prs = Presentation()
        self.prs.slide_height = self.height
        self.prs.slide_width = self.width

    def get_coordinate(self, rel_coordinate):
        rel_left, rel_top, rel_width, rel_height = rel_coordinate
        left = self.get_absolute_cordinate(length=self.width, rel_ratio=rel_left)
        top = self.get_absolute_cordinate(length=self.height, rel_ratio=rel_top)
        width = self.get_absolute_cordinate(length=self.width, rel_ratio=rel_width)
        height = self.get_absolute_cordinate(length=self.height, rel_ratio=rel_height)
        return left, top, width, height
    
    def create_slide(self, slide_idx=6):
        """
            0: "Title (presentation title slide)",
            1: "Title and Content",
            2: "Section Header (sometimes called Segue)",
            3: "Two Content (side by side bullet textboxes)",
            4: "Comparison (same but additional title for each side by side content box)",
            5: "Title Only",
            6: "Blank",
            7: "Content with Caption",
            8: "Picture with Caption"
        """ 
        slide_layout = self.prs.slide_layouts[slide_idx]
        return self.prs.slides.add_slide(slide_layout)

    def save_pptx(self, filename=None, save_path=None):
        if save_path is None:
            save_path = pathlib.Path.home().joinpath("Downloads")
        elif not isinstance(save_path, pathlib.WindowsPath):
            try:
                save_path = pathlib.Path(save_path)
            except Exception as e:
                raise ValueError(f"存储路径错误，{e}")

            if not save_path.is_dir():
                raise ValueError(f"{save_path}不是目录")

        if filename is None:
            filename = "output.pptx"

        self.prs.save(save_path.joinpath(filename))

    def add_textbox(
            self, 
            cur_slide: slide.Slide, 
            rel_coordinate=(0, 0, 1, 1), 
            paragraph_infos=None,
            vertical_anchor = MSO_ANCHOR.MIDDLE
        ):
        left, top, width, height = self.get_coordinate(rel_coordinate=rel_coordinate)
        shape = cur_slide.shapes.add_textbox(
            left, top, width, height
        )

        text_frame = shape.text_frame
        # 自动换行
        text_frame.word_wrap = True
        # 垂直居中, TOP MIDDLE BOTTOM
        text_frame.vertical_anchor = vertical_anchor
        if paragraph_infos is not None:
            self.add_text(text_frame=text_frame, paragraph_infos=paragraph_infos)
    
    def add_text(self, text_frame, paragraph_infos):
        first_paragraph = True
        for paragraph_info in paragraph_infos:
            if first_paragraph:
                p = text_frame.paragraphs[0]
                first_paragraph = False
            else:
                p = text_frame.add_paragraph()
            
            p.alignment = PARAGRAPH_ALIGN.get(paragraph_info.get("对齐", "左对齐"), PP_ALIGN.LEFT)
            p.space_after = Pt(paragraph_info.get("段落间距", DEFAULT_PARAGRAPH_SPACING))
            p.level = paragraph_info.get("level", 0)

            texts_infos = paragraph_info.get("文字", [])
            if texts_infos:
                for text, font_info in texts_infos:
                    run = p.add_run()
                    run.text = text
                    set_font(run.font, font_info)
    
    def add_rectangle_shape(
        self, 
        cur_slide: slide.Slide, 
        rel_coordinate=(0, 0, 1, 1), 
        paragraph_infos=None,
        vertical_anchor=MSO_ANCHOR.TOP,
        background_color=(255, 255, 255),
        line_color=(255, 255, 255),
        rounded=False
    ):  
        """可以用下面这个代码看形状的具体样子
                import pathlib
                from pptx import Presentation
                from pptx.util import Inches
                from pptx.enum.shapes import MSO_SHAPE

                prs = Presentation()
                slide_layout = prs.slide_layouts[6]
                for shape_name in [i for i in dir(MSO_SHAPE) if not i.startswith("_")]:
                    slide = prs.slides.add_slide(slide_layout)
                    left, top, width, height = Inches(1), Inches(1), Inches(3), Inches(3)
                    shape_obj = getattr(MSO_SHAPE, shape_name)
                    try:
                        shape = slide.shapes.add_shape(
                            shape_obj, left, top, width, height
                        )
                        text_frame = shape.text_frame
                        p = text_frame.paragraphs[0]
                        run = p.add_run()
                        run.text = shape_name
                    except Exception as e:
                        # 删除这个slide
                        xml_slides = prs.slides._sldIdLst  # pylint: disable=W0212
                        slides = list(xml_slides)
                        xml_slides.remove(slides[-1])
                save_path = pathlib.Path.home().joinpath("Downloads")
                prs.save(save_path.joinpath("shapes.pptx"))
        """
        left, top, width, height = self.get_coordinate(rel_coordinate=rel_coordinate)
        shape_name = MSO_SHAPE.ROUNDED_RECTANGLE if rounded else MSO_SHAPE.RECTANGLE
        shape = cur_slide.shapes.add_shape(
            shape_name, left, top, width, height
        )
        # 填充颜色
        shape.fill.solid()
        shape.fill.fore_color.rgb = RGBColor(*background_color)
        # 轮廓颜色
        line = shape.line
        line.color.rgb = RGBColor(*line_color)
        # 取消阴影
        shape.shadow.inherit = False

        # 填充文字
        text_frame = shape.text_frame
        # 自动换行
        text_frame.word_wrap = True
        # 垂直居中, TOP MIDDLE BOTTOM
        text_frame.vertical_anchor = vertical_anchor
        if paragraph_infos is not None:
            self.add_text(text_frame=text_frame, paragraph_infos=paragraph_infos)


    def add_picture(self, cur_slide: slide.Slide, image_path, rel_coordinate=(0, 0, 1, 1)):
        left, top, width, height = self.get_coordinate(rel_coordinate=rel_coordinate)
        _ = cur_slide.shapes.add_picture(
            image_file=image_path, left=left, top=top, width=width, height=height
        )
    
    def add_bar_chart(
        self,
        cur_slide: slide.Slide, 
        chart_data: CategoryChartData,
        rel_coordinate=(0, 0, 1, 1),
        bar_type="纵向",
        title=None,
        title_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
        series_colors=None,
        tick_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
        rotation_x=None,
        rotation_y=None,
        data_labels_font_info=None,
        data_labels_position=XL_LABEL_POSITION.INSIDE_END,
        data_labels_format=None
    ):
        ChartCreator().create_bar_chart(
            cur_slide=cur_slide, 
            chart_data=chart_data,
            absolute_coordinate=self.get_coordinate(rel_coordinate=rel_coordinate),
            bar_type=bar_type,
            title=title,
            title_font_info=title_font_info,
            series_colors=series_colors,
            tick_font_info=tick_font_info,
            rotation_x=rotation_x,
            rotation_y=rotation_y,
            data_labels_font_info=data_labels_font_info,
            data_labels_position=data_labels_position,
            data_labels_format=data_labels_format
        )

    def add_pie_chart(
            self, 
            cur_slide: slide.Slide, 
            chart_data: CategoryChartData,
            rel_coordinate=(0, 0, 1, 1),
            title=None,
            title_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
            series_colors=None,
            data_labels_font_info=None,
            data_labels_position=XL_LABEL_POSITION.OUTSIDE_END,
            legend_font_info=None,
            legend_position=XL_LEGEND_POSITION.BOTTOM, 
            data_labels_format="0.00%"
        ):
        ChartCreator().create_pie_chart(
            cur_slide=cur_slide, 
            chart_data=chart_data,
            absolute_coordinate=self.get_coordinate(rel_coordinate=rel_coordinate),
            title=title,
            title_font_info=title_font_info,
            series_colors=series_colors,
            data_labels_font_info=data_labels_font_info,
            data_labels_position=data_labels_position,
            legend_font_info=legend_font_info,
            legend_position=legend_position, 
            data_labels_format=data_labels_format
        )
    
    def add_line_chart(
        self,
        cur_slide: slide.Slide,
        chart_data: CategoryChartData,
        rel_coordinate=(0, 0, 1, 1),
        title='标题',
        title_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
        legend_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
        legend_position=XL_LEGEND_POSITION.BOTTOM,
        series_colors=None,
        marker_colors=None,
        marker_style=XL_MARKER_STYLE.CIRCLE,
        smooth=False,
        tick_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
        axis_color=(0, 0, 0),
        rotation_x=None,
        rotation_y=None,
        data_labels_font_info=None,
        data_labels_position=XL_LABEL_POSITION.INSIDE_END
    ):
        ChartCreator().create_line_chart(
            cur_slide=cur_slide, 
            chart_data=chart_data,
            absolute_coordinate=self.get_coordinate(rel_coordinate=rel_coordinate),
            title=title,
            title_font_info=title_font_info,
            legend_font_info=legend_font_info,
            legend_position=legend_position,
            series_colors=series_colors,
            marker_colors=marker_colors,
            marker_style=marker_style,
            smooth=smooth,
            tick_font_info=tick_font_info,
            axis_color=axis_color,
            rotation_x=rotation_x,
            rotation_y=rotation_y,
            data_labels_font_info=data_labels_font_info,
            data_labels_position=data_labels_position
        )

    def add_table(
        self, 
        cur_slide: slide.Slide, 
        data: pd.DataFrame, 
        rel_coordinate=(0, 0, 1, 1),
        col_widths=None,
        row_heights=None,
        header_font_info=("微软雅黑", 14, True, False, (255, 255, 255)),
        header_fore_color=(79, 128, 189),
        odd_cell_fore_color=(233, 237, 244),
        even_cell_fore_color=(255, 255, 255),
        cell_font_info=("微软雅黑", 12, False, False, (0, 0, 0))
    ):
        num_rows = data.shape[0] + 1
        num_cols = data.shape[1]
        
        # 表格大小
        left, top, width, height = self.get_coordinate(rel_coordinate=rel_coordinate)
        shape = cur_slide.shapes.add_table(rows=num_rows, cols=num_cols, left=left, top=top, width=width, height=height)
        table = shape.table

        # 列宽
        if col_widths is None:
            col_widths = [int(1 / num_cols * 100) / 100] * num_cols
            col_widths[-1] = 1 - sum(col_widths[:-1])
        for cur_col_idx, width_ratio in zip(range(num_cols), col_widths):
            table.columns[cur_col_idx].width = self.get_absolute_cordinate(length=width, rel_ratio=width_ratio)

        # 行高
        if row_heights is None:
            row_heights =[int(1 / num_rows * 100) / 100] * num_rows
            row_heights[-1] = 1 - sum(row_heights[:-1])
        for cur_row_idx, height_ratio in zip(range(num_rows), row_heights):
            table.rows[cur_row_idx].height = self.get_absolute_cordinate(length=height, rel_ratio=height_ratio)

        # 设置值和格式
        data_with_header = [data.columns.tolist()] + data.values.tolist()
        for cur_row_idx, cur_col_idx in product(range(num_rows), range(num_cols)):
            
            cur_cell = table.cell(cur_row_idx, cur_col_idx)
            cur_cell.text = f"{data_with_header[cur_row_idx][cur_col_idx]}"

            if cur_row_idx == 0:
                cur_cell_color = header_fore_color
                cur_cell_font_info = header_font_info
            else:
                cur_cell_font_info = cell_font_info
                cur_cell_color = odd_cell_fore_color if cur_row_idx % 2 == 1 else even_cell_fore_color
            set_color(cur_cell, cur_cell_color)
            
            cur_text_frame = cur_cell.text_frame
            cur_paragraph = cur_text_frame.paragraphs[0]
            set_font(
                cur_font=cur_paragraph.font,
                cur_font_info=cur_cell_font_info
            )
            # 水平居中
            cur_paragraph.alignment = PARAGRAPH_ALIGN["居中"]
            # 自动换行
            cur_text_frame.word_wrap = True
            # 垂直居中
            cur_text_frame.vertical_anchor = MSO_ANCHOR.MIDDLE



from pptx.util import Pt
from docx.oxml.ns import qn
from pptx.dml.color import RGBColor


def set_font(cur_font, cur_font_info):
    """设置字体
    Args:
        cur_font (_type_): pptx.text.text.font类型
        cur_font_info (_type_): tuple, 字体、字号、加粗、斜体、颜色
    """
    font_name, font_size, bold, italic, rgb = cur_font_info

    cur_font.name = font_name
    # 下面这坨是因为python-pptx包不怎么支持中文字体，不管咋设置都是宋体
    if cur_font._rPr.find(qn("a:ea")) is not None:
        cur_font._rPr.find(qn("a:ea")).set("typeface", font_name)
    else:
        element = cur_font._rPr.makeelement(qn("a:ea"))
        element.set("typeface", font_name)
        cur_font._rPr.append(element)
    
    cur_font.size = Pt(font_size)
    cur_font.bold = bold
    cur_font.italic = italic
    cur_font.color.rgb = RGBColor(*rgb)


def set_color(obj, color):
    """设置对象颜色

    Args:
        obj (_type_): series.format.line, series.marker.format, font等任意类型
        color (_type_): RGB 三元组tuple
    """
    obj.fill.solid()
    obj.fill.fore_color.rgb = RGBColor(*color)



from pptx.enum.text import PP_ALIGN


DEFAULT_PARAGRAPH_SPACING = 8


PARAGRAPH_ALIGN = {
    "居中": PP_ALIGN.CENTER,
    "左对齐": PP_ALIGN.LEFT,
    "右对齐": PP_ALIGN.RIGHT,
    # 均匀分布
    "均匀": PP_ALIGN.DISTRIBUTE,
    
    # 每行都在边距处开始和结束，并调整了单词之间的间距, 这样一来，该行就填满了段落的宽度。
    "填满": PP_ALIGN.JUSTIFY,
    # 在单词之间使用少量空格
    '填满2': PP_ALIGN.JUSTIFY_LOW,
    "THAI_DISTRIBUTE": PP_ALIGN.THAI_DISTRIBUTE
}





import warnings

from pptx import slide
from pptx.util import Inches
from pptx.chart.chart import Chart
from pptx.chart.data import CategoryChartData 
from pptx.enum.chart import XL_CHART_TYPE, XL_LEGEND_POSITION, XL_MARKER_STYLE, XL_LABEL_POSITION

from automate_office.automate_ppt.helper_func import set_font, set_color, RGBColor


class ChartCreator(object):

    @staticmethod
    def set_title(cur_chart: Chart, title='标题', cur_font_info=("微软雅黑", 18, False, False, (89, 89, 89))):
        cur_chart.has_title = True
        cur_chart.chart_title.has_text_frame = True
        title_paragraph = cur_chart.chart_title.text_frame.paragraphs[0]
        title_paragraph.text = title
        set_font(
            cur_font=title_paragraph.font,
            cur_font_info=cur_font_info
        )
    
    @staticmethod
    def set_legend(
        cur_chart: Chart, 
        legend_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
        legend_position=XL_LEGEND_POSITION.BOTTOM
    ):
        # 设置图例
        cur_chart.has_legend = True
        cur_chart.legend.position = legend_position
        cur_chart.legend.include_in_layout = False
        set_font(
            cur_font=cur_chart.legend.font,
            cur_font_info=legend_font_info
        )

    @staticmethod
    def set_series_color(cur_chart: Chart, series_colors, color_type='self'):
        """

        Args:
            cur_chart (Chart): _description_
            series_colors (_type_): _description_
            color_type (str, optional): 在折线图里，这个就是line
        """
        try:
            plot = cur_chart.plots[0]
            for series, color in zip(plot.series, series_colors):
                if color_type != "self":
                    color_obj = getattr(series.format, color_type)
                else:
                    color_obj = series.format
                set_color(color_obj, color)
        except Exception as e:
            warnings.warn(f"系列颜色设置失败, {e}")


    @staticmethod
    def set_marker(cur_chart: Chart, marker_colors, marker_style=XL_MARKER_STYLE.CIRCLE):
        try:
            plot = cur_chart.plots[0]
            for i, series in enumerate(plot.series):
                # 设置marker样式
                series.marker.style = marker_style

                if marker_colors is not None and i < len(marker_colors):
                    color = marker_colors[i]
                    # marker颜色
                    set_color(obj=series.marker.format, color=color)
                    # marker轮廓颜色，不设置的话就是默认颜色，和设置的颜色不一样了
                    set_color(obj=series.marker.format.line, color=color)
             
        except Exception as e:
            warnings.warn(f"图例样式设置失败, {e}")

    @staticmethod
    def set_tick(cur_chart: Chart, tick_name, tick_font_info, axis_color=(0, 0, 0), rotation=None):
        if tick_name == "x":
            # 横坐标轴
            cur_axis = cur_chart.category_axis
            # tick_labels = cur_chart.category_axis.tick_labels
        elif tick_name == "y":
            # 纵坐标轴
            cur_axis = cur_chart.value_axis
            # tick_labels = cur_chart.value_axis.tick_labels
        else:
            raise ValueError(f"轴名称出错, 输入的是{tick_name}")

        set_font(
            cur_font=cur_axis.tick_labels.font,
            cur_font_info=tick_font_info
        )
        cur_chart.category_axis.format.line.color.rgb = RGBColor(*axis_color)

        if rotation is not None:
            txPr = cur_axis.tick_labels._element.get_or_add_txPr()
            # X轴标签旋转45度
            # 在 OpenXML SDK 里面，采用的基础单位是 60000 倍的 Degree 角度值
            # 也就是在获取到 OpenXML 的 Int32Value 时，获取数值，除以 60000 就拿到了角度值
            # 例如： 逆时针旋转45度，那么就等于 60000 * (-45) = -2700000
            txPr.bodyPr.set("rot", f"{60000 * rotation:>.0f}")

    @staticmethod
    def set_data_labels(cur_chart: Chart, data_labels_font_info, data_labels_position=XL_LABEL_POSITION.INSIDE_END, data_labels_format=None):
        plot = cur_chart.plots[0]
        plot.has_data_labels = True

        for series in cur_chart.plots[0].series:
            data_labels = series.data_labels
            set_font(data_labels.font, data_labels_font_info)
            data_labels.position = data_labels_position
            data_labels.show_value = True
            if data_labels_format is not None:
                data_labels.number_format = data_labels_format

    def create_line_chart(
            self,
            cur_slide: slide.Slide, 
            chart_data: CategoryChartData,
            absolute_coordinate=(Inches(0), Inches(0), Inches(7.5), Inches(7.5)),
            title=None,
            title_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
            legend_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
            legend_position=XL_LEGEND_POSITION.BOTTOM,
            series_colors=None,
            marker_colors=None,
            marker_style=XL_MARKER_STYLE.CIRCLE,
            smooth=False,
            tick_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
            axis_color=(0, 0, 0),
            rotation_x=None,
            rotation_y=None,
            data_labels_font_info=None,
            data_labels_position=XL_LABEL_POSITION.INSIDE_END
        ):
        x, y, cx, cy = absolute_coordinate
        graphic_frame = cur_slide.shapes.add_chart(
            XL_CHART_TYPE.LINE, x, y, cx, cy, chart_data
        )
        chart = graphic_frame.chart

        # 设置标题
        if title is not None:
            self.set_title(cur_chart=chart, title=title, cur_font_info=title_font_info)
        # 设置图例
        self.set_legend(cur_chart=chart, legend_font_info=legend_font_info, legend_position=legend_position)
        # 设置系列颜色
        if series_colors is not None:
            self.set_series_color(
                cur_chart=chart, 
                series_colors=series_colors, 
                color_type='line'
            )
        # 设置点样式
        self.set_marker(
            cur_chart=chart, 
            marker_colors=marker_colors, 
            marker_style=marker_style
        )

        # 去掉网格
        chart.value_axis.has_major_gridlines = False

        # 平滑折线图
        for series in chart.series:
            series.smooth = smooth

        # 设定x轴y轴
        self.set_tick(cur_chart=chart, tick_name="x", tick_font_info=tick_font_info, axis_color=axis_color, rotation=rotation_x)
        self.set_tick(cur_chart=chart, tick_name="y", tick_font_info=tick_font_info, axis_color=axis_color, rotation=rotation_y)

        # 设定数据标注data_labels
        if data_labels_font_info is not None:
            self.set_data_labels(cur_chart=chart, data_labels_font_info=data_labels_font_info, data_labels_position=data_labels_position)

    def create_bar_chart(
            self,
            cur_slide: slide.Slide, 
            chart_data: CategoryChartData,
            absolute_coordinate=(Inches(0), Inches(0), Inches(7.5), Inches(7.5)),
            bar_type="纵向",
            title=None,
            title_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
            series_colors=None,
            tick_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
            axis_color=(0, 0, 0),
            rotation_x=None,
            rotation_y=None,
            data_labels_font_info=None,
            data_labels_position=XL_LABEL_POSITION.INSIDE_END, 
            data_labels_format=None
        ):
        x, y, cx, cy = absolute_coordinate
        if bar_type == "纵向":
            chart_name = XL_CHART_TYPE.COLUMN_CLUSTERED
        elif bar_type == "横向":
            chart_name = XL_CHART_TYPE.BAR_CLUSTERED
        else:
            chart_name = XL_CHART_TYPE.COLUMN_CLUSTERED
        graphic_frame = cur_slide.shapes.add_chart(
            chart_name, x, y, cx, cy, chart_data
        )
        chart = graphic_frame.chart

        # 设置标题
        if title is not None:
            self.set_title(cur_chart=chart, title=title, cur_font_info=title_font_info)

        # 去掉网格线
        chart.value_axis.has_major_gridlines = False

        # 设置系列颜色
        if series_colors is not None:
            self.set_series_color(
                cur_chart=chart, 
                series_colors=series_colors, 
                color_type="self"
            )

        # 设定x轴y轴
        self.set_tick(cur_chart=chart, tick_name="x", tick_font_info=tick_font_info, axis_color=axis_color, rotation=rotation_x)
        self.set_tick(cur_chart=chart, tick_name="y", tick_font_info=tick_font_info, axis_color=axis_color, rotation=rotation_y)

        # 设定数据标注data_labels
        if data_labels_font_info is not None:
            self.set_data_labels(cur_chart=chart, data_labels_font_info=data_labels_font_info, data_labels_position=data_labels_position, data_labels_format=data_labels_format)

    def create_pie_chart(
            self,
            cur_slide: slide.Slide, 
            chart_data: CategoryChartData,
            absolute_coordinate=(Inches(0), Inches(0), Inches(7.5), Inches(7.5)),
            title=None,
            title_font_info=("微软雅黑", 18, False, False, (89, 89, 89)),
            series_colors=None,
            data_labels_font_info=None,
            data_labels_position=XL_LABEL_POSITION.OUTSIDE_END,
            legend_font_info=None,
            legend_position=XL_LEGEND_POSITION.BOTTOM, 
            data_labels_format="0.00%"
        ):

        x, y, cx, cy = absolute_coordinate

        graphic_frame = cur_slide.shapes.add_chart(
            XL_CHART_TYPE.PIE, x, y, cx, cy, chart_data
        )
        chart = graphic_frame.chart

        # 设置标题
        if title is not None:
            self.set_title(cur_chart=chart, title=title, cur_font_info=title_font_info)

        # # 去掉网格线
        # chart.value_axis.has_major_gridlines = False

        # 设置系列颜色
        if series_colors is not None:
            for point, color in zip(chart.series[0].points, series_colors):
                set_color(point.format, color)

        # 设定数据标注data_labels
        if data_labels_font_info is not None:
            self.set_data_labels(cur_chart=chart, data_labels_font_info=data_labels_font_info, data_labels_position=data_labels_position, data_labels_format=data_labels_format)

        # 设置图例
        if legend_font_info is not None:
            chart.has_legend = True
            chart.legend.include_in_layout = False
            self.set_legend(cur_chart=chart, legend_font_info=legend_font_info, legend_position=legend_position)



       




from pptx.util import Pt
from pptx.enum.text import PP_ALIGN, MSO_ANCHOR

from automate_office.automate_ppt.ppt_enums import PARAGRAPH_ALIGN, DEFAULT_PARAGRAPH_SPACING
from automate_office.automate_ppt.helper_func import set_font


def set_textframe(text_frame, paragraph_infos=None, vertical_anchor = MSO_ANCHOR.MIDDLE):
    # text_frame.clear()

    # 自动换行
    text_frame.word_wrap = True
    # 垂直居中, TOP MIDDLE BOTTOM
    text_frame.vertical_anchor = vertical_anchor

    for idx, cur_paragraph_info in enumerate(paragraph_infos):
        cur_paragraph = text_frame.paragraphs[idx]
        cur_paragraph.clear()

        cur_paragraph.alignment = PARAGRAPH_ALIGN.get(cur_paragraph_info.get("对齐", "左对齐"), PP_ALIGN.LEFT)
        cur_paragraph.space_after = Pt(cur_paragraph_info.get("段落间距", DEFAULT_PARAGRAPH_SPACING))
        cur_paragraph.level = cur_paragraph_info.get("level", 0)

        texts_infos = cur_paragraph_info.get("文字", [])
        if texts_infos:
            for text, font_info in texts_infos:
                run = cur_paragraph.add_run()
                run.text = text
                set_font(run.font, font_info)


def replace_chart_data(chart, chart_data):
    chart.replace_data(chart_data)



"""
    未完成，需要时再处理，无非是字体、段落、页眉页脚、表格、列表等的设置
"""

import pathlib

from docx.oxml.ns import qn
from docx import Document, section
from docx.shared import Inches, Cm
from docx.enum.section import WD_SECTION_START, WD_ORIENT

from automate_office.automate_word.helper_func import add_paragraph


class WordCreator(object):

    def __init__(self, section_config={}):
        self._create_doc(section_config=section_config)

    def _create_doc(self, section_config):
        self.doc =  Document()
        if section_config:
            self.set_section(cur_section=self.doc.sections[0], cur_section_config=section_config)
            
    
    def save_doc(self, filename=None, save_path=None):
        if save_path is None:
            save_path = pathlib.Path.home().joinpath("Downloads")
        elif not isinstance(save_path, pathlib.WindowsPath):
            try:
                save_path = pathlib.Path(save_path)
            except Exception as e:
                raise ValueError(f"存储路径错误，{e}")

            if not save_path.is_dir():
                raise ValueError(f"{save_path}不是目录")

        if filename is None:
            filename = "output.docx"

        self.doc.save(save_path.joinpath(filename))

    def set_section(
        self, 
        cur_section: section.Section, 
        cur_section_config={
            "unit_type": "inches",
            "params": {
                "top_margin": 1,
                "bottom_margin": 1,
                "left_margin": 1.25,
                "right_margin": 1.25,
                "page_height": 11,
                "page_width": 8.5,
            },
            # 目前不管用
            "orientation": WD_ORIENT.PORTRAIT
        }
    ):
        support_params = ("top_margin", "bottom_margin", "left_margin", "right_margin", "page_height", "page_width")

        unit_type = {
            "inches": Inches,
            "cm": Cm
        }.get(
            cur_section_config.get("unit_type", "inches"), Inches
        )
        
        setattr(cur_section, "orientation", cur_section_config.get("orientation", WD_ORIENT.PORTRAIT))

        for k, v in cur_section_config["params"].items():
            if k in support_params:
                setattr(cur_section, k, unit_type(v))

    def create_section(self, section_config={}, start_type=WD_SECTION_START.NEW_PAGE):
        new_section = self.doc.add_section(start_type=start_type)
        if section_config:
            self.set_section(cur_section=new_section, cur_section_config=section_config)
        return new_section

    @staticmethod
    def set_multiple_col(cur_section: section.Section, n_cols):
        """分栏: 最多支持45
        """
        cur_section._sectPr.xpath('./w:cols')[0].set(qn('w:num'), f'{n_cols:>.0f}') 
    
    def add_header(self, text, level=0):
        self.doc.add_heading(text, level)

    def add_paragraph(self, paragraph_info, paragraph_style=None):
  
        paragraph = self.doc.add_paragraph(style=paragraph_style)
        add_paragraph(cur_paragraph=paragraph, paragraph_info=paragraph_info)


    def 页眉页脚():
        from docx import Document
        document = Document()  # 新建文档  或 模板
        header = document.sections[0].header  # 获取第一个节的页眉
        print('页眉中默认段落数：', len(header.paragraphs))
        paragraph = header.paragraphs[0]  # 获取页眉的第一个段落
        paragraph.add_run('这是第一节的页眉2')  # 添加页面内容
        footer = document.sections[0].footer  # 获取第一个节的页脚
        paragraph = footer.paragraphs[0]  # 获取页脚的第一个段落
        paragraph.add_run('这是第一节的页脚2')  # 添加页脚内容
        document.save('test.docx')  # 保存文档

    def about_list(self):
        
        for i, s in enumerate(
            [
                "List", "List 2", "List 3", 
                "List Bullet", "List Bullet 2", "List Bullet 3", 
                "List Continue", "List Continue 2", "List Continue 3", 
                "List Number", "List Number 2", "List Number 3"
            ]
        ):
            _ = self.doc.add_paragraph(str(i) * 10, style=s)

        self.save_doc()


from setuptools import setup, find_packages


setup(
    name="automate_office",
    version="0.3.0",
    packages=find_packages(),
    install_requires=[
        # 在这里列出你的库所需的其他Python包
        "python-docx==1.1.0",
        "python-pptx==0.6.23",
        "openpyxl==3.1.2"
    ],

    author="Sun Meng",
    author_email="clivesun@163.com",
    description="Offic自动化工具-自用，在原有包的基础上二次开发, PPT部分暂时告一段落, Word暂时用处不大, Excel做了一些函数和例子",
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    license="",
    url="",
    classifiers=[
        # 发展时期,常见的如下
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # 开发的目标用户
        "Intended Audience :: Developers",
        # 属于什么类型
        "Topic :: Software Development :: Build Tools",
        # 许可证信息
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
)


import pathlib
import datetime

import numpy as np
import pandas as pd

from automate_office.automate_excel.excel_funcs import *


def get_border_style(cur_row_idx, cur_col_idx,row_idx_range=(1, 100), col_idx_range=(1, 100), border_color="00FF6238"):
    left_style, right_style, top_style, bottom_style = ["dashed"] * 4
    if cur_row_idx == row_idx_range[0]:
        top_style = "thin"
    if cur_row_idx == row_idx_range[1]:
        bottom_style = "thin"
    if cur_col_idx == col_idx_range[0]:
        left_style = "thin"
    if cur_col_idx == col_idx_range[1]:
        right_style = "thin"
    return create_border_style(right_style=right_style, left_style=left_style, bottom_style=bottom_style, top_style=top_style, color=border_color)


def get_number_format(cur_row_idx, cur_col_idx):
    if cur_row_idx == 1:
        return "General"
    
    if cur_col_idx == 2:
        return "0.00%"
    elif cur_col_idx == 3:
        return "#,##0.00;[红色]-#,##0.00"
    elif cur_col_idx == 4:
        # 中文在自适应列宽的时候似乎总会短一截
        return 'yyyy"年"m"月"d"日"'
    else:
        return "General"
    

if __name__ == "__main__":  
    file_path = pathlib.Path.home().joinpath("Downloads/format.xlsx")
    sheet_name = "Format"

    df = pd.DataFrame(
        {
            "col1": list("abcdefg"),
            "col2": np.random.random(size=(7, )),
            "col3": np.random.randint(low=-1e5, high=1e5, size=(7, )),
            "col4": pd.date_range(start=datetime.datetime(2023, 12, 1).date(), end=datetime.datetime(2023, 12, 7).date())
        }
    )
    df.to_excel(file_path, index=False, sheet_name=sheet_name)
    n_rows, n_cols = df.shape[0] + 1, df.shape[1]

    # 格式化
    wb = open_excel(filename=file_path)
    ws = wb[sheet_name]

    alignment = creaete_alignment_style(horizontal="center", vertical="center")
    title_font = create_font_style(font_name="微软雅黑", font_size=16, font_color="00FFFFFF", bold=True)
    content_font = create_font_style(font_name="微软雅黑", font_size=12, font_color="00000000")
    title_fill = create_fill_style(fill_type="solid", start_color="00FF6238")

    # 遍历单元格设置格式
    for row_cells in ws.iter_rows():
        for cur_cell in row_cells:
            if cur_cell.row == 1:
                cur_cell.font = title_font
                cur_cell.fill = title_fill
            else:
                cur_cell.font = content_font

            cur_cell.alignment = alignment
            cur_cell.border = get_border_style(col_idx_range=(1, n_cols), row_idx_range=(1, n_rows), cur_col_idx=cur_cell.column, cur_row_idx=cur_cell.row)
            cur_cell.number_format = get_number_format(cur_col_idx=cur_cell.column, cur_row_idx=cur_cell.row)

    # 自适应列宽
    get_autojust_col_width(ws, n_cols)

    wb.save(file_path)
